from __future__ import annotations

import logging
from dataclasses import dataclass
from datetime import timedelta
from enum import IntEnum
from typing import ClassVar, Self

from .exceptions import DecodeError

_PACKET_REGISTRY: dict[int, list[type[PacketNotify]]] = {}
_LOGGER = logging.getLogger(__name__)


def from_scaled_nullable(data: bytes, scale: float) -> float | None:
    if all(x == 0xFF for x in data):
        return None
    return int.from_bytes(data, "big") / scale


def to_scaled_nullable(data: float | None, length: int, scale: float) -> bytes:
    if data is None:
        return bytes([0xFF] * length)
    return round(data * scale).to_bytes(length, "big")


def from_nullable(data: bytes) -> int | None:
    if all(x == 0xFF for x in data):
        return None
    return int.from_bytes(data, "big")


def from_nullable_enum(data: bytes, enum: type[IntEnum]) -> int | None:
    if all(x == 0xFF for x in data):
        return None
    value = int.from_bytes(data, "big")
    try:
        return enum(value)
    except ValueError:
        return value


def to_nullable(data: int | None, length: int) -> bytes:
    if data is None:
        return bytes([0xFF] * length)
    return data.to_bytes(length, "big")


@dataclass
class Packet:
    type: ClassVar[int]

    @classmethod
    def decode(cls, data: bytes) -> Self:
        raise NotImplementedError()

    def encode(self) -> bytes:
        raise NotImplementedError()


@dataclass
class PacketNotify(Packet):
    def __init_subclass__(cls, /, **kwargs):
        super().__init_subclass__(**kwargs)
        if packet_type := getattr(cls, "type", None):
            _PACKET_REGISTRY.setdefault(packet_type, []).append(cls)

    @classmethod
    def decode(cls, data: bytes) -> PacketNotify:
        if len(data) < 1:
            raise DecodeError("Failed to parse packet")
        exceptions = []
        for registered_cls in _PACKET_REGISTRY.get(data[0], []):
            try:
                return registered_cls.decode(data)
            except DecodeError as exc:
                exceptions.append(exc)
        if exceptions:
            raise ExceptionGroup(f"Fail to decode {data}", exceptions)
        return PacketUnknown(data[0], data[1:])

    @classmethod
    def request(cls) -> bytes:
        raise NotImplementedError


@dataclass
class PacketNotifyAck(PacketNotify):
    """Set timer."""

    type: ClassVar[int]
    data: int

    @classmethod
    def decode(cls, data: bytes) -> Self:
        if len(data) < 2:
            raise DecodeError("Packet too short")
        return cls(data=data[1])


@dataclass
class PacketWrite(Packet):
    """Base class fro packet writes."""


@dataclass
class PacketA0Notify(PacketNotify):
    """Device status"""

    type: ClassVar[int] = 0xA0
    battery: int
    version_major: int
    version_minor: int
    function_type: int
    probe_count: int
    ambient: bool
    alarm_interval: int
    alarm_sound: bool

    @classmethod
    def decode(cls, data: bytes) -> Self:
        if len(data) < 6:
            raise DecodeError("Packet too short")
        if data[0] != cls.type:
            raise DecodeError("Failed to parse packet")

        battery = data[1]
        version_major = data[2]
        version_minor = data[3]
        _unknown = data[4]
        bitfield = data[5]
        function_type = bitfield & 0xF
        probe_count = (bitfield >> 4) & 0x7
        ambient = bool(bitfield >> 7)

        alarm_interval = 5
        alarm_sound = True
        if len(data) > 6:
            alarm_interval = data[6]
            alarm_sound = data[7] == 1

        return cls(
            battery=battery,
            version_major=version_major,
            version_minor=version_minor,
            function_type=function_type,
            probe_count=probe_count,
            ambient=ambient,
            alarm_interval=alarm_interval,
            alarm_sound=alarm_sound,
        )

    @classmethod
    def request(cls) -> bytes:
        return bytes(
            [
                cls.type,
                0x00,
                0x00,
            ]
        )


@dataclass
class PacketA1Notify(PacketNotify):
    """Temperature on probes"""

    type: ClassVar[int] = 0xA1
    temperatures: list[float | None]

    @classmethod
    def decode(cls, data: bytes) -> Self:
        if len(data) < 1:
            raise DecodeError("Packet too short")
        if data[0] != cls.type:
            raise DecodeError("Failed to parse packet")

        temperatures = [
            int.from_bytes(data[index : index + 2], "big") for index in range(1, len(data), 2)
        ]

        def convert(value: int) -> float | None:
            if value == 65535:
                return None
            if value > 32768:
                return (value - 32768) / 10
            return value / 10

        temperatures = [convert(temperature) for temperature in temperatures]

        return cls(temperatures=temperatures)

    @classmethod
    def request(cls) -> bytes:
        return bytes(
            [
                cls.type,
                0x00,
            ]
        )


@dataclass
class PacketA3Notify(PacketNotifyAck):
    type: ClassVar[int] = 0xA3


@dataclass
class PacketA300Write(PacketWrite):
    """Set min max temperature."""

    type: ClassVar[int] = 0xA3
    alarm_type: ClassVar[int] = 0x00
    probe: int
    minimum: float | None
    maximum: float | None

    @classmethod
    def decode(cls, data: bytes) -> Self:
        if len(data) < 7:
            raise DecodeError("Packet too short")
        if data[2] != cls.alarm_type:
            raise DecodeError("Invalid subtype")
        return cls(
            probe=data[1],
            minimum=from_scaled_nullable(data[3:5], 10.0),
            maximum=from_scaled_nullable(data[5:7], 10.0),
        )

    def encode(self) -> bytes:
        return bytes(
            [
                self.type,
                self.probe,
                self.alarm_type,
                *to_scaled_nullable(self.minimum, 2, 10.0),
                *to_scaled_nullable(self.maximum, 2, 10.0),
            ]
        )


@dataclass(kw_only=True)
class PacketA301Write(PacketWrite):
    """Set target temperature."""

    type: ClassVar[int] = 0xA3
    alarm_type: ClassVar[int] = 0x01
    probe: int
    target: float | None

    @classmethod
    def decode(cls, data: bytes) -> Self:
        if len(data) < 7:
            raise DecodeError("Packet too short")
        if data[2] != cls.alarm_type:
            raise DecodeError("Invalid subtype")

        return cls(
            probe=data[1],
            target=from_scaled_nullable(data[3:5], 10.0),
        )

    def encode(self) -> bytes:
        return bytes(
            [
                self.type,
                self.probe,
                self.alarm_type,
                *to_scaled_nullable(self.target, 2, 10.0),
                0,
                0,
            ]
        )


@dataclass(kw_only=True)
class PacketA303Write(PacketWrite):
    """Set target temperature."""

    type: ClassVar[int] = 0xA3
    alarm_type: ClassVar[int] = 0x03
    probe: int
    grill_type: int

    @classmethod
    def decode(cls, data: bytes) -> Self:
        if len(data) < 7:
            raise DecodeError("Packet too short")
        if data[2] != cls.alarm_type:
            raise DecodeError("Invalid subtype")
        return cls(
            probe=data[1],
            grill_type=data[4],
        )

    def encode(self) -> bytes:
        return bytes([self.type, self.probe, self.alarm_type, 0, self.grill_type, 0, 0])


@dataclass
class PacketA5Notify(PacketNotify):
    """Status from probe"""

    type: ClassVar[int] = 0xA5
    probe: int
    message: int

    class Message(IntEnum):
        PROBE_ACKNOWLEDGE = 0
        PROBE_ALARM = 5
        PROBE_DISCONNECTED = 6

    @classmethod
    def decode(cls, data: bytes) -> Self:
        if len(data) < 3:
            raise DecodeError("Packet too short")
        if data[0] != cls.type:
            raise DecodeError("Failed to parse packet")

        try:
            message = PacketA5Notify.Message(data[2])
        except ValueError:
            message = data[2]

        return cls(probe=data[1], message=message)


@dataclass
class PacketA6Write(PacketWrite):
    """Set alarm behaviour."""

    class Unit(IntEnum):
        UNIT_CELCIUS = 0
        UNIT_FARENHEIT = 1

    type: ClassVar[int] = 0xA6
    temperature_unit: int | None = None
    alarm_interval: int | None = None

    @classmethod
    def decode(cls, data: bytes) -> Self:
        if len(data) < 3:
            raise DecodeError("Packet too short")
        if data[1] == 0xFF:
            temperature_unit = None
        else:
            try:
                temperature_unit = PacketA6Write.Unit(data[1])
            except ValueError:
                temperature_unit = data[2]

        if data[2] == 0xFF:
            alarm_interval = None
        else:
            alarm_interval = data[2]

        return cls(
            temperature_unit=temperature_unit,
            alarm_interval=alarm_interval,
        )

    def encode(self) -> bytes:
        if self.temperature_unit is None:
            temperature_unit_data = 0xFF
        else:
            temperature_unit_data = self.temperature_unit

        if self.alarm_interval is None:
            alarm_interval_data = 0xFF
        else:
            alarm_interval_data = self.alarm_interval

        return bytes(
            [
                self.type,
                temperature_unit_data,
                alarm_interval_data,
            ]
        )


@dataclass
class PacketA7Write(PacketWrite):
    """Set timer."""

    type: ClassVar[int] = 0xA7
    probe: int
    time: timedelta
    unknown: int = 1

    @classmethod
    def decode(cls, data: bytes) -> Self:
        if len(data) < 5:
            raise DecodeError("Packet too short")
        return cls(
            time=timedelta(seconds=int.from_bytes(data[3:5], "big")), probe=data[1], unknown=data[2]
        )

    def encode(self) -> bytes:
        seconds = round(self.time.total_seconds())
        return bytes(
            [
                self.type,
                self.probe,
                self.unknown,
                *seconds.to_bytes(2, "big"),
            ]
        )


@dataclass
class PacketA7Notify(PacketNotifyAck):
    type: ClassVar[int] = 0xA7


@dataclass
class PacketA8Write(PacketWrite):
    """Set alarm behaviour."""

    type: ClassVar[int] = 0xA8
    probe: int
    unknown: int = 0

    @classmethod
    def decode(cls, data: bytes) -> Self:
        if len(data) < 3:
            raise DecodeError("Packet too short")
        if data[0] != cls.type:
            raise DecodeError("Invalid type")
        return cls(probe=data[1], unknown=data[2])

    def encode(self) -> bytes:
        return bytes(
            [
                self.type,
                self.probe,
                self.unknown,
            ]
        )


@dataclass
class PacketA8Notify(PacketNotify):
    """Status from probe"""

    class AlarmType(IntEnum):
        TEMPERATURE_TARGET = 0
        TEMPERATURE_RANGE = 1

    type: ClassVar[int] = 0xA8
    probe: int
    alarm_type: int | None
    temperature_1: float | None = None
    temperature_2: float | None = None
    grill_type: int = 0
    time: timedelta = timedelta()
    unknown_1: bytes = bytes([0x00])
    unknown_2: bytes = bytes([0, 0])

    @classmethod
    def decode(cls, data: bytes) -> Self:
        if len(data) < 12:
            raise DecodeError("Packet too short")
        if data[0] != cls.type:
            raise DecodeError("Failed to parse type")

        return cls(
            probe=data[1],
            alarm_type=from_nullable_enum(data[2:3], PacketA8Notify.AlarmType),
            temperature_1=from_scaled_nullable(data[3:5], 10.0),
            temperature_2=from_scaled_nullable(data[5:7], 10.0),
            unknown_1=data[7:8],
            grill_type=data[8],
            unknown_2=data[9:11],
            time=timedelta(seconds=int.from_bytes(data[11:13], "big")),
        )

    def encode(self) -> bytes:
        seconds = round(self.time.total_seconds())
        return bytes(
            [
                self.type,
                self.probe,
                *to_nullable(self.alarm_type, 1),
                *to_scaled_nullable(self.temperature_1, 2, 10.0),
                *to_scaled_nullable(self.temperature_2, 2, 10.0),
                *self.unknown_1,
                self.grill_type,
                *self.unknown_2,
                *seconds.to_bytes(2, "big"),
            ]
        )


@dataclass
class PacketUnknown(PacketNotify):
    type: int
    data: bytes

    @classmethod
    def decode(cls, data: bytes) -> Self:
        if len(data) < 1:
            raise DecodeError("Packet too short")
        return cls(data[0], data=data[1:])



from datetime import timedelta

import anyio
import asyncclick as click
from bleak import (
    BleakClient,
    BleakScanner,
)
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData
from bleak.uuids import uuidstr_to_str

from .const import MainService, ManufacturerData
from .exceptions import DecodeError
from .packets import (
    PacketA0Notify,
    PacketA1Notify,
    PacketA7Write,
    PacketA300Write,
    PacketA301Write,
    PacketNotify,
)
from .services import Characteristic, NotifyCharacteristic, WriteCharacteristic


@click.group()
async def cli():
    pass


@cli.command()
async def scan():
    click.echo("Scanning for devices")

    devices = set()

    def detected(device: BLEDevice, advertisement: AdvertisementData):
        if device not in devices:
            if MainService.uuid not in advertisement.service_uuids:
                return
            devices.add(device)

        click.echo(f"Device: {device}")
        for service in advertisement.service_uuids:
            click.echo(f" - Service: {service} {uuidstr_to_str(service)}")
        click.echo(f" - Data: {advertisement.service_data}")
        click.echo(f" - Manu: {advertisement.manufacturer_data}")

        if data := advertisement.manufacturer_data.get(ManufacturerData.company):
            decoded = ManufacturerData.decode(data)
            click.echo(f" -     : {decoded}")

        click.echo(f" - RSSI: {advertisement.rssi}")
        click.echo()

    async with BleakScanner(detected, service_uuids=[MainService.uuid]):
        await anyio.sleep_forever()


@cli.group(chain=True)
@click.argument("address")
@click.option("--code", default="")
@click.pass_context
async def connect(ctx: click.Context, address: str, code: str):
    click.echo(f"Connecting to: {address} ...", nl=False)
    client = await ctx.with_async_resource(BleakClient(address, timeout=20))
    ctx.obj = client
    click.echo(" Done")

    def notify_data(char_specifier: BleakGATTCharacteristic, data: bytearray):
        try:
            packet_data = NotifyCharacteristic.decode(data)
            packet = PacketNotify.decode(packet_data)
            click.echo(f"Notify: {packet}")
        except DecodeError as exc:
            click.echo(f"Failed to decode: {data.hex()} with error {exc}")

    await client.start_notify(MainService.notify.uuid, notify_data)

    await client.write_gatt_char(
        MainService.write.uuid, WriteCharacteristic.encode(PacketA0Notify.request()), False
    )
    await client.write_gatt_char(
        MainService.write.uuid, WriteCharacteristic.encode(PacketA1Notify.request()), False
    )


@connect.command()
@click.pass_obj
async def list(client: BleakClient):
    for service in client.services:
        click.echo(f"Service: {service}")

        async def read_print(char: BleakGATTCharacteristic):
            parser = Characteristic.registry.get(char.uuid)
            if "read" in char.properties:
                data = await client.read_gatt_char(char.uuid)
            else:
                data = None
            click.echo(f" -  {char}")
            click.echo(f" -  {char.properties}")
            if data is not None and parser:
                click.echo(f" -  Data: {parser.decode(data)}")

        async with anyio.create_task_group() as tg:
            for char in service.characteristics:
                tg.start_soon(read_print, char)


@connect.command()
@click.argument("probe", type=int)
@click.argument("seconds", type=int)
@click.pass_obj
async def timer(client: BleakClient, probe: int, seconds: int):
    click.echo(f"Setting timer on {probe} for delay {seconds} ...", nl=False)
    await client.write_gatt_char(
        MainService.write.uuid,
        WriteCharacteristic.encode(
            PacketA7Write(probe=probe, time=timedelta(seconds=seconds)).encode()
        ),
        False,
    )
    click.echo(" Done")


@connect.command()
@click.argument("probe", type=int)
@click.argument("minimum", type=float)
@click.argument("maximum", type=float)
@click.pass_obj
async def range(client: BleakClient, probe: int, minimum: float, maximum: float):
    click.echo(f"Setting range cook on {probe} min: {minimum} max: {maximum} ...", nl=False)
    await client.write_gatt_char(
        MainService.write.uuid,
        WriteCharacteristic.encode(
            PacketA300Write(probe=probe, minimum=minimum, maximum=maximum).encode()
        ),
        False,
    )
    click.echo(" Done")


@connect.command()
@click.argument("probe", type=int)
@click.argument("target", type=float)
@click.pass_obj
async def target(client: BleakClient, probe: int, target: float):
    click.echo(f"Setting range cook on {probe} target: {target} ...", nl=False)
    await client.write_gatt_char(
        MainService.write.uuid,
        WriteCharacteristic.encode(PacketA301Write(probe=probe, target=target).encode()),
        False,
    )
    click.echo(" Done")


@connect.command()
@click.argument("data", type=str)
@click.pass_obj
async def request(client: BleakClient, data: str):
    click.echo(f"Sending request : {data} ...", nl=False)
    data_raw = bytes.fromhex(data)
    await client.write_gatt_char(
        MainService.write.uuid,
        WriteCharacteristic.encode(data_raw),
        False,
    )
    click.echo(" Done")


@connect.command()
async def wait():
    click.echo("Waiting")
    await anyio.sleep_forever()


def main():
    try:
        cli()
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()



"""Constants"""

from dataclasses import dataclass
from typing import ClassVar

from togrill_bluetooth.services import NotifyCharacteristic, WriteCharacteristic

from .services import Service


class MainService(Service):
    uuid = "0000cee0-0000-1000-8000-00805f9b34fb"
    write = WriteCharacteristic()
    notify = NotifyCharacteristic()


@dataclass
class ManufacturerData:
    company: ClassVar[int] = 0x879A

    @staticmethod
    def decode(data: bytes):
        return ManufacturerData()



from __future__ import annotations

import logging
from asyncio import Future
from collections.abc import Callable
from typing import TypeVar

from bleak import BleakClient
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.device import BLEDevice
from bleak_retry_connector import establish_connection

from .const import MainService
from .exceptions import DecodeError, WriteFailed
from .packets import Packet, PacketNotify, PacketNotifyAck, PacketWrite
from .services import NotifyCharacteristic, WriteCharacteristic

_LOGGER = logging.getLogger(__name__)

_PacketNotifyType = TypeVar("_PacketNotifyType", bound=PacketNotify)


class Client:
    def __init__(
        self, client: BleakClient, notify_callback: Callable[[Packet], None] | None
    ) -> None:
        self.bleak_client = client
        self._notify_callbacks = []
        if notify_callback:
            self._notify_callbacks.append(notify_callback)

    @property
    def is_connected(self) -> bool:
        return self.bleak_client.is_connected

    def notify_callbacks(self, packet: Packet):
        for callback in self._notify_callbacks:
            callback(packet)

    async def _start_notify(self):
        def notify_data(char_specifier: BleakGATTCharacteristic, data: bytearray):
            try:
                packet_data = NotifyCharacteristic.decode(data)
                packet = PacketNotify.decode(packet_data)
                _LOGGER.debug("Notify: %s", packet)
            except DecodeError as exc:
                _LOGGER.error("Failed to decode: %s with error %s", data, exc)
            self.notify_callbacks(packet)

        await self.bleak_client.start_notify(MainService.notify.uuid, notify_data)

    @staticmethod
    async def connect(
        device: BLEDevice,
        notify_callback: Callable[[Packet], None] | None = None,
        disconnected_callback: Callable[[], None] | None = None,
    ) -> Client:
        def _disconnected_callback(client: BleakClient):
            _LOGGER.info("Device disconnected %s", client.address)
            if disconnected_callback:
                disconnected_callback()

        bleak_client = await establish_connection(
            BleakClient,
            device=device,
            name="ToGrill Connection",
            disconnected_callback=_disconnected_callback,
        )
        try:
            client = Client(bleak_client, notify_callback)
            await client._start_notify()
        except Exception:
            await bleak_client.disconnect()
            pass
        return client

    async def disconnect(self) -> None:
        await self.bleak_client.disconnect()

    async def request(self, packet: type[PacketNotify]) -> None:
        await self.bleak_client.write_gatt_char(
            MainService.write.uuid, WriteCharacteristic.encode(packet.request()), False
        )

    async def write(self, packet: PacketWrite) -> PacketNotify:
        result_future = Future[PacketNotify]()

        def _callback(packet_notify: Packet):
            if not isinstance(packet_notify, PacketNotify):
                return

            if packet.type != packet.type:
                return

            if result_future.cancelled() or result_future.done():
                return

            if isinstance(packet_notify, PacketNotifyAck):
                if packet_notify.data == 0:
                    result_future.set_exception(WriteFailed("Failed to write data"))
                    return

            result_future.set_result(packet_notify)

        self._notify_callbacks.append(_callback)
        try:
            await self.bleak_client.write_gatt_char(
                MainService.write.uuid, WriteCharacteristic.encode(packet.encode()), False
            )
            return await result_future
        finally:
            self._notify_callbacks.remove(_callback)

    async def read(self, packet_type: type[_PacketNotifyType]) -> _PacketNotifyType:
        result = Future[packet_type]()

        def _callback(packet: Packet):
            if isinstance(packet, packet_type):
                if not result.cancelled() and not result.done():
                    result.set_result(packet)

        self._notify_callbacks.append(_callback)
        try:
            await self.request(packet_type)
            return await result
        finally:
            self._notify_callbacks.remove(_callback)



from __future__ import annotations

from dataclasses import dataclass
from functools import reduce
from itertools import chain, tee
from operator import xor
from typing import ClassVar, Generic, TypeVar

from .exceptions import DecodeError

CharacteristicType = TypeVar("CharacteristicType")

_PAYLOAD_PREFIX = [0x55, 0xAA]


def pretty_name(name: str):
    data = name.split("_")
    return " ".join(f"{part[0].upper()}{part[1:]}" for part in data)


def wrap_payload(data: bytes) -> bytes:
    """Wraps the payload with the prefix and checksum."""
    payload = chain(_PAYLOAD_PREFIX, len(data).to_bytes(2, "big"), data)
    payload, payload_copy = tee(payload, 2)
    checksum = reduce(xor, payload_copy)
    return bytes(chain(payload, [checksum]))


def unwrap_payload(data: bytes) -> bytes:
    """Unwraps the payload, removing the prefix and checksum."""
    if len(data) < 5 or data[0:2] != bytes(_PAYLOAD_PREFIX):
        raise DecodeError("Invalid payload format")

    payload_len = int.from_bytes(data[2:4], "big")
    if len(data) - 5 != payload_len:
        raise DecodeError("Payload size mismatch")

    payload = data[4 : 4 + payload_len]
    checksum = data[4 + payload_len]
    checksum_expected = reduce(xor, data[:-1])

    if checksum_expected != checksum:
        raise DecodeError(f"Expected checksum {checksum_expected:x} found {checksum:x}")

    return payload


@dataclass
class Characteristic(Generic[CharacteristicType]):
    uuid: str
    name: str = ""
    registry: ClassVar[dict[str, Characteristic]] = {}

    def __set_name__(self, _, name: str):
        self.name = pretty_name(name)

    def __post_init__(self):
        self.registry[self.uuid] = self

    @classmethod
    def decode(cls, data: bytes) -> CharacteristicType:
        raise NotImplementedError(f"Decoding of {type(cls)} is not implemented")

    @classmethod
    def encode(cls, data: CharacteristicType) -> bytes:
        raise NotImplementedError(f"Encoding of {type(cls)} is not implemented")


class Service:
    uuid: ClassVar[str]
    registry: ClassVar[dict[str, type[Service]]] = {}

    def __init_subclass__(cls, /, **kwargs):
        super().__init_subclass__(**kwargs)
        cls.registry[cls.uuid] = cls

    @classmethod
    def characteristics(cls):
        for value in vars(cls).values():
            if isinstance(value, Characteristic):
                yield value


@dataclass
class NotifyCharacteristic(Characteristic[bytes]):
    uuid: ClassVar[str] = "0000cee2-0000-1000-8000-00805f9b34fb"

    @staticmethod
    def decode(data: bytes) -> bytes:
        return unwrap_payload(data)

    @staticmethod
    def encode(data: bytes) -> bytes:
        return wrap_payload(data)


@dataclass
class WriteCharacteristic(Characteristic[bytes]):
    uuid: ClassVar[str] = "0000cee1-0000-1000-8000-00805f9b34fb"

    @staticmethod
    def decode(data: bytes) -> bytes:
        return unwrap_payload(data)

    @staticmethod
    def encode(data: bytes) -> bytes:
        return wrap_payload(data)



import pytest

from togrill_bluetooth.services import (
    NotifyCharacteristic,
    WriteCharacteristic,
)


@pytest.mark.parametrize(
    "data,result", [("A100", "55AA0002A1005C"), ("A00000", "55AA0003A000005C")]
)
def test_encode_write(data, result):
    assert (
        WriteCharacteristic.encode(bytes.fromhex(data)).hex() == bytes(bytes.fromhex(result)).hex()
    )


@pytest.mark.parametrize(
    "data,result",
    [
        ("55aa0008a05b00080060050160", "a05b000800600501"),
        ("55aa000fa1ffffffffffffffffffffffffffff51", "a1ffffffffffffffffffffffffffff"),
    ],
)
def test_decode_notify(data, result):
    assert (
        NotifyCharacteristic.decode(bytes.fromhex(data)).hex() == bytes(bytes.fromhex(result)).hex()
    )



from unittest.mock import Mock

from bleak import BleakClient

from togrill_bluetooth.client import Client
from togrill_bluetooth.packets import PacketUnknown


def test_client_notify():
    mock_ble_client = Mock(spec_set=BleakClient)

    callback = Mock()
    packet = PacketUnknown(0, b"")

    client = Client(mock_ble_client, callback)
    client.notify_callbacks(packet)

    callback.assert_called_once_with(packet)



from datetime import timedelta

import pytest

from togrill_bluetooth.packets import (
    Packet,
    PacketA0Notify,
    PacketA1Notify,
    PacketA6Write,
    PacketA7Write,
    PacketA8Notify,
    PacketA300Write,
    PacketA301Write,
    PacketA303Write,
    PacketNotify,
    PacketUnknown,
)


@pytest.mark.parametrize(
    "data,result",
    [
        (
            "a05b000800600501",
            PacketA0Notify(
                battery=91,
                version_major=0,
                version_minor=8,
                function_type=0,
                probe_count=6,
                ambient=False,
                alarm_interval=5,
                alarm_sound=True,
            ),
        ),
        (
            "a1ffffffffffffffffffffffffffff",
            PacketA1Notify(temperatures=[None, None, None, None, None, None, None]),
        ),
        (
            "a1 ffff ffff ffff ffff ffff ffff 01b5",
            PacketA1Notify(temperatures=[None, None, None, None, None, None, 43.7]),
        ),
        (
            "a8 01 00 03 e8 ff ff 00 05 00 00 00 00",
            PacketA8Notify(
                probe=1, alarm_type=0, temperature_1=100, temperature_2=None, grill_type=5
            ),
        ),
        (
            "a8 02 00 03 e8 03 e9 00 05 00 00 00 00",
            PacketA8Notify(
                probe=2, alarm_type=0, temperature_1=100, temperature_2=100.1, grill_type=5
            ),
        ),
        (
            "a8 01 01 03e8 ffff 00 05 00 00 00 00",
            PacketA8Notify(probe=1, alarm_type=1, temperature_1=100, grill_type=5),
        ),
        (
            "a8 01 ff 03e8 ffff 00 05 00 00 00 00",
            PacketA8Notify(probe=1, alarm_type=None, temperature_1=100, grill_type=5),
        ),
        (
            "00ffffffffffffffffffffffffffff",
            PacketUnknown(0x00, bytes.fromhex("ffffffffffffffffffffffffffff")),
        ),
    ],
)
def test_decode_packet(data: str, result: Packet):
    packet = PacketNotify.decode(bytes.fromhex(data))
    assert packet == result


@pytest.mark.parametrize(
    "packet,raw",
    [
        (
            PacketA7Write(probe=0, time=timedelta(seconds=16), unknown=1),
            "a700010010",
        ),
        (
            PacketA6Write(temperature_unit=PacketA6Write.Unit.UNIT_CELCIUS, alarm_interval=5),
            "a60005",
        ),
        (
            PacketA6Write(temperature_unit=PacketA6Write.Unit.UNIT_FARENHEIT),
            "a601ff",
        ),
        (
            PacketA6Write(alarm_interval=15),
            "a6ff0f",
        ),
        (
            PacketA7Write(probe=0, time=timedelta(seconds=256), unknown=1),
            "a700010100",
        ),
        (
            PacketA300Write(probe=1, minimum=1.6, maximum=3.2),
            "a3 01 00 0010 0020",
        ),
        (
            PacketA301Write(probe=1, target=1.6),
            "a3 01 01 0010 0000",
        ),
        (
            PacketA303Write(probe=1, grill_type=5),
            "a3 01 03 0005 0000",
        ),
    ],
)
def test_roundtrip_packet(packet: Packet, raw: str):
    raw_bytes = bytes.fromhex(raw)
    packet_bytes = packet.encode()
    assert packet_bytes == raw_bytes
    assert packet == packet.decode(raw_bytes)



"""ToGrill main module"""

from bleak.uuids import register_uuids

from .services import Service

SUPPORTED_DEVICES = {
    "Pro-00",
    "Pro-10",
    "Mini5",
    "MiniClk",
    "MW-01",
    "NEX-01",
    "NEX-02",
    "AT-01",
    "AT-02",
    "Nano",
    "HTD",
    "AP-MTT980",
    "WM-01",
    "NC01",
    "Candy",
    "Pro-01",
    "Pro-02",
    "Pro-03",
    "Pro-04",
    "Pro-05",
    "Pro-06",
    "Pro-07",
    "Pro-08",
    "Pro-09",
    "Pro10",
    "X1",
    "X2",
    "Nano2",
    "Kweight",
    "WPprobe",
    "WPpCHAR",
    "WP-01",
    "WP-02",
    "WP02",
    "WP03",
    "WP03A",
    "WP03B",
    "KT",
    "KT1",
    "KT2",
    "HB1",
    "HB2",
    "DM1",
    "ED1",
    "EN1",
    "EN2",
    "YFC1",
    "YFC2",
    "WPpbcon",
    "WPpbCHG",
    "WPR01",
    "WPR02",
    "WPR03",
    "NPL1",
    "SG1",
    "WPS2",
    "PS2",
    "WPS1",
    "PS1",
    "ES1",
    "SW1",
    "MAY",
    "IGNITE",
    "WP05",
    "WP06",
    "WP07",
    "WP08",
    "WP08N",
    "FC1",
    "FC2",
    "GP1",
    "HL02",
    "TXD01",
    "BI01",
    "BI02",
    "ACCU-PRO",
    "ATK-02",
    "GM1",
    "NanoS",
    "BP55",
    "BP05",
    "BP40",
    "AT-02W1",
    "AT-02W2",
    "AT-02W4",
    "BI01W1",
    "BI01W2",
    "BI01W4",
    "OG2",
    "ALFA01H",
    "AT-03",
}

register_uuids(
    {service.uuid: f"ToGrill {service.__name__}" for service in Service.registry.values()}
)

register_uuids(
    {
        char.uuid: f"ToGrill {service.__name__} {char.name}"
        for service in Service.registry.values()
        for char in service.characteristics()
    }
)
